import os
import logging
import subprocess
import tempfile
import glob
import rpm
import shutil
import re

from shlex import quote
from munch import Munch
from tempfile import NamedTemporaryFile

from .utils import run, QUIET, MODEST, VERBOSE
from .errors import RpkgError

log = logging.getLogger('__main__')

class RpmPackage(object):

    def __init__(self, pkg_path, spec_content=None, spec_path=None, outdir=None,
                 use_existing_outdir=False, preprocess_spec=True,
                 base_output_path='/tmp/rpkg',
                 default_macros=[], user_macros=[],
                 bcond_with=[], bcond_without=[],
                 rpmdefines=[], bump_rpmdefines=[],
                 rpmlint_config=None, copr_project=None,
                 produce_sources=True, version_bump=False,
                 release_bump=False, git_props=Munch(),
                 verbosity=MODEST):

        self.pkg_path = pkg_path
        self.git_props = git_props
        self.use_existing_outdir = use_existing_outdir
        self.preprocess_spec = preprocess_spec
        self.base_output_path = base_output_path
        self.user_macros = user_macros
        self.bcond_with = bcond_with
        self.bcond_without = bcond_without
        self.rpmdefines = rpmdefines
        self.bump_rpmdefines = bump_rpmdefines
        self.rpmlint_config = rpmlint_config
        self.produce_sources = produce_sources
        self.version_bump = version_bump
        self.release_bump = release_bump
        self.default_macros = default_macros
        self.verbosity = verbosity
        self.copr_project = copr_project
        self._spec_content = spec_content
        self._spec_path = spec_path
        self._outdir = outdir
        self.create_base_outdir()

    ######################## properties ########################

    @property
    def spec_content(self):
        if not self._spec_content:
            with open(self.spec_path, 'r') as f:
                self._spec_content = f.read()
        return self._spec_content

    @property
    def spec_path(self):
        if not self._spec_path:
            self._spec_path = self.locate_spec(self.pkg_path)
        return self._spec_path

    @property
    def outdir(self):
        if not self._outdir:
            self._outdir = self.create_outdir()
        return self._outdir

    @spec_content.setter
    def spec_content(self, value):
        self._spec_content = value

    @spec_path.setter
    def spec_path(self, value):
        self._spec_path = value

    @outdir.setter
    def outdir(self, value):
        self._outdir = value

    ######################## helpers ########################

    @classmethod
    def locate_spec(cls, path):
        spec_path = None

        rpkg_ext_first = lambda f: (0 if f.endswith('.rpkg') else 1, f)

        for f in sorted(os.listdir(path), key=rpkg_ext_first):
            if (f.endswith('.spec.rpkg') or f.endswith('.spec')) and not f.startswith('.'):
                spec_path = os.path.abspath(os.path.join(path, f))
                break

        if not spec_path:
            raise RpkgError('No spec file found at {0}.'
                            .format(path))

        return spec_path

    def create_base_outdir(self):
        if not os.path.isdir(self.base_output_path):
            prev_umask = os.umask(0)
            os.makedirs(self.base_output_path, mode=0o1777)
            os.umask(prev_umask)

    def create_outdir(self):
        dirname_base = os.path.splitext(
            os.path.splitext(os.path.basename(self.spec_path))[0])[0]

        cnt = len(glob.glob(os.path.join(self.base_output_path, dirname_base+'*')))

        if not self.use_existing_outdir:
            prefix = dirname_base + '-' + str(cnt+1) + '-'
            return tempfile.mkdtemp(dir=self.base_output_path, prefix=prefix)

        try:
            prefix = dirname_base + '-' + str(cnt) + '-'
            return glob.glob(os.path.join(self.base_output_path, prefix+'*'))[0]
        except IndexError:
            raise RpkgError('Could not load the latest outdir with prefix "%s" at %s'
                            % (prefix, self.base_output_path))

    def load_rpm_config(self):
        rpm.addMacro('_sourcedir', self.pkg_path)

        if self.version_bump or self.release_bump:
            for rpmdefine in self.bump_rpmdefines:
                name, expr = rpmdefine.split()
                rpm.addMacro(name, expr)
        else:
            for rpmdefine in self.rpmdefines:
                name, expr = rpmdefine.split()
                rpm.addMacro(name, expr)

    def rpm_opts(self):
        rpm_opts = []

        if self.version_bump or self.release_bump:
            for rpmdefine in self.bump_rpmdefines:
                rpm_opts.extend(['--define', rpmdefine])
        else:
            for rpmdefine in self.rpmdefines:
                rpm_opts.extend(['--define', rpmdefine])

        return rpm_opts

    def rpm_parse_opts(self):
        return [ '--define', '_sourcedir %s' % self.pkg_path ] + self.rpm_opts()

    def rpmbuild_opts(self):
        rpmbuild_opts = [
            '--define', '_sourcedir %s' % self.outdir,
            '--define', '_specdir %s' % self.outdir,
            '--define', '_builddir %s' % self.outdir,
            '--define', '_buildrootdir %s' % self.outdir,
            '--define', '_srcrpmdir %s' % self.outdir,
            '--define', '_rpmdir %s' % self.outdir,
        ]

        for arg in self.bcond_with:
            rpmbuild_opts.extend(['--with', arg])

        for arg in self.bcond_without:
            rpmbuild_opts.extend(['--without', arg])

        rpmbuild_opts.extend(self.rpm_opts())
        return rpmbuild_opts

    def nvrid(self, name=None, version=None, release=None):
        if name and version and release:
            return '-'.join([name, version, release])

        nvre = self.nvre()
        name = name or nvre.name
        version = version or nvre.version
        release = release or nvre.release
        return '-'.join([name, version, release])

    def nvre(self):
        """Get name, version, release, epoch of the package."""
        cmd = ['rpm'] + self.rpm_parse_opts()
        final_tmp_spec_path = self.final_tmp_spec()

        # We put the question tag delimiters into our query so that
        # we can split the output when there are multiple subpackages.
        cmd.extend(['-q', '--qf', '%{NAME} %{VERSION} %{RELEASE} %{EPOCH}??',
                    '--specfile', '%s' % final_tmp_spec_path])

        output = run(cmd, capture=True).out
        os.unlink(final_tmp_spec_path)

        # Split output by ??, grab the first part and
        # split again to get name, epoch, ver, and rel.
        first_line_output = output.split('??')[0]
        parts = first_line_output.split()

        if len(parts) != 4:
            raise RpkgError('Could not get n-v-r-e from %r'
                            % first_line_output)

        return Munch({
            'name': parts[0],
            'version': parts[1],
            'release': parts[2],
            'epoch': parts[3],
        })

    def preproc_source_params(self):
        for default_macro_file in self.default_macros:
            source_params = '-s {0} '.format(os.path.join(
                '/usr/lib/rpkg.macros.d', default_macro_file))

        for user_macro_file in self.user_macros:
            if not os.path.exists(user_macro_file):
                raise RpkgError('{0} does not exist.'.format(user_macro_file))
            source_params += '-s {0} '.format(quote(user_macro_file))

        return source_params.strip()

    def preproc_env_params(self, spec_templ):
        env_params = '-e INPUT_PATH={0}'.format(quote(spec_templ))

        if self.produce_sources:
            env_params += ' -e OUTDIR={0}'.format(quote(self.outdir))

        if self.version_bump:
            env_params += ' -e VERSION_BUMP=1'

        if self.release_bump:
            env_params += ' -e RELEASE_BUMP=1'

        if self.verbosity == VERBOSE:
            env_params += ' -e VERBOSE=1'

        return env_params

    def preproc_git_props_params(self, status_file,
                                 merged_tag_refs_file,
                                 submodule_refs_file):
        env_params = ''

        merged_tag_refs_file.write(self.git_props.merged_tag_refs+'\n')
        submodule_refs_file.write(self.git_props.submodule_refs+'\n')
        status_file.write(self.git_props.status+'\n')

        merged_tag_refs_file.close()
        submodule_refs_file.close()
        status_file.close()

        for key, val in self.git_props.items():
            if key == 'merged_tag_refs':
                env_params += ' -e GIT_MERGED_TAG_REFS_FILE={0}'.format(
                    quote(merged_tag_refs_file.name))
            elif key == 'submodule_refs':
                env_params += ' -e GIT_SUBMODULE_REFS_FILE={0}'.format(
                    quote(submodule_refs_file.name))
            elif key == 'status':
                env_params += ' -e GIT_STATUS_FILE={0}'.format(
                    quote(status_file.name))
            else:
                env_params += ' -e GIT_{0}={1}'.format(
                    key.upper(), quote(val))

        return env_params.strip()

    def final_spec_text(self):
        if not self.preprocess_spec:
            return self.spec_content

        source_params = self.preproc_source_params()
        env_params = self.preproc_env_params(self.spec_path)

        tmp0_file = NamedTemporaryFile(
            'w', prefix='tmp_', dir=self.base_output_path, delete=False)
        tmp1_file = NamedTemporaryFile(
            'w', prefix='tmp_', dir=self.base_output_path, delete=False)
        tmp2_file = NamedTemporaryFile(
            'w', prefix='tmp_', dir=self.base_output_path, delete=False)
        tmp3_file = NamedTemporaryFile(
            'w', prefix='tmp_', dir=self.base_output_path, delete=False)

        git_props_params = self.preproc_git_props_params(
            tmp1_file, tmp2_file, tmp3_file)

        tmp0_file.write(self.spec_content)
        tmp0_file.close()

        cmd = 'cat {0} | preproc -C {1} {2} {3} {4}'.format(
            quote(tmp0_file.name), quote(os.getcwd()), source_params,
            env_params, git_props_params)

        try:
            final_spec_text = run(cmd, shell=True, capture=True).out
        finally:
            if not self.verbosity == VERBOSE:
                os.unlink(tmp0_file.name)
                os.unlink(tmp1_file.name)
                os.unlink(tmp2_file.name)
                os.unlink(tmp3_file.name)

        return final_spec_text

    def final_spec_path(self):
        final_spec_basename = os.path.basename(
            self.spec_path[:-5] if self.spec_path.endswith('.rpkg') else self.spec_path)
        return os.path.join(self.outdir, final_spec_basename)

    def final_spec(self):
        final_spec_path = self.final_spec_path()

        with open(final_spec_path, 'w') as f:
            f.write(self.final_spec_text()+'\n')

        if self.produce_sources:
            self.setup_source_symlinks(final_spec_path)

        log.info('Wrote: {0}'.format(final_spec_path))
        return final_spec_path

    def final_tmp_spec(self):
        final_tmp_spec = NamedTemporaryFile(
            'w', prefix='tmp_',
            dir=self.base_output_path,
            delete=False)

        self.produce_sources = False
        final_tmp_spec.write(self.final_spec_text())
        final_tmp_spec.close()

        return final_tmp_spec.name

    def setup_source_symlinks(self, final_spec_path):
        ts = rpm.ts()

        try:
            self.load_rpm_config()
            rpm_spec = ts.parseSpec(final_spec_path)
        except ValueError as e:
            raise RpkgError("Could not parse spec file with error: "+str(e))
        finally:
            rpm.reloadConfig()

        for (filename, num, flags) in rpm_spec.sources:
            filename = os.path.basename(filename)
            symlink_source = os.path.join(self.pkg_path, filename)
            symlink_dest = os.path.join(self.outdir, filename)

            if os.path.isfile(symlink_source) and not \
                    os.path.isfile(symlink_dest):
                os.symlink(symlink_source, symlink_dest)

    ######################## commands ########################

    def clog(self, raw=False):
        cmd = ['rpm'] + self.rpm_parse_opts()

        final_tmp_spec_path = self.final_tmp_spec()
        cmd.extend(['-q', '--qf', '%{CHANGELOGTEXT}??',
                    '--specfile', '%s' % final_tmp_spec_path])

        output = run(cmd, capture=True).out
        os.unlink(final_tmp_spec_path)

        # get output for first subpackage
        first_pkg_output = output.split('??')[0].strip()

        if raw:
            return first_pkg_output

        return re.sub(r'^-?\s*(.*)$', r'\1',
                      first_pkg_output, flags=re.M)

    def compile(self, target=None, short=False, nocheck=False,
                bcond_with=[], bcond_without=[]):
        """Run rpmbuild -bc on the package

        Optionally for a specific target, short-circuit it, or
        specify with/without options.

        Logs the output and returns nothing.
        """
        cmd = ['rpmbuild'] + self.rpmbuild_opts()

        if target:
            cmd.extend(['--target', target])

        if short:
            cmd.extend(['--short-circuit'])

        if nocheck:
            cmd.extend(['--nocheck'])

        if self.verbosity == QUIET:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-bc', self.final_spec()])
        run(cmd, capture=False)

    def copr_build(self, project, nowait, config_file):
        self.srpm()
        try:
            srpm_path = glob.glob(
                os.path.join(self.outdir, '*.src.rpm'))[0]
        except:
            raise RpkgError('Failed to generate srpm.')

        cmd = ['copr-cli']

        if config_file:
            cmd.extend(['--config', config_file])

        cmd.extend(['build'])

        if nowait:
            cmd.extend(['--nowait'])

        project = project or self.copr_project
        if not project:
            raise RpkgError('Copr project name not specified.')

        cmd.extend([project, srpm_path])
        run(cmd, capture=False)

    def install(self, target=None, short=False, nocheck=False,
                bcond_with=[], bcond_without=[]):
        """Run rpm -bi on the package

        Optionally for a specific target, short-circuit it, or
        specify with/without options

        Logs the output and returns nothing
        """
        cmd = ['rpmbuild'] + self.rpmbuild_opts()

        if target:
            cmd.extend(['--target', target])

        if short:
            cmd.extend(['--short-circuit'])

        if nocheck:
            cmd.extend(['--nocheck'])

        if self.verbosity == QUIET:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-bi', self.final_spec()])
        run(cmd, capture=False)

    def lint(self, info=False, rpmlint_config=None, target_path=None):
        """Run rpmlint over built files

        Logs the output and returns nothing
        """
        if not target_path:
            self.use_existing_outdir = True
            target_path = self.outdir

        log.info('Working on {0}...'.format(target_path))

        cmd = ['rpmlint']

        if info:
            cmd.extend(['-i'])

        if rpmlint_config:
            cmd.extend(["-f", rpmlint_config])
        elif self.rpmlint_config:
            cmd.extend(["-f", self.rpmlint_config])

        cmd.extend(['--'])
        cmd.extend([target_path])
        run(cmd, capture=False)

    def local(self, target=None, nocheck=False, bcond_with=[], bcond_without=[]):
        """rpmbuild locally for given target.

        Writes output to a log file and logs it to the logger
        Returns nothing
        """
        cmd = ['rpmbuild'] + self.rpmbuild_opts()

        if target:
            cmd.extend(['--target', target])

        if nocheck:
            cmd.extend(['--nocheck'])

        if self.verbosity == QUIET:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['-ba', self.final_spec()])
        run(cmd, capture=False)

    def prep(self, target=None, bcond_with=[], bcond_without=[]):
        """Run rpm -bp on the package

        Optionally for a specific target, short-circuit it, or
        specify with/without options.

        Logs the output and returns nothing.
        """
        cmd = ['rpmbuild'] + self.rpmbuild_opts()

        if target:
            cmd.extend(['--target', target])

        if self.verbosity == QUIET:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['--nodeps', '-bp', self.final_spec()])
        run(cmd, capture=False)

    def srpm(self, target=None, bcond_with=[], bcond_without=[]):
        """
        Create an srpm using hashtype from content in the package
        Requires sources already downloaded.
        """
        cmd = ['rpmbuild'] + self.rpmbuild_opts()

        if target:
            cmd.extend(['--target', target])

        if self.verbosity == QUIET:
            cmd.extend(['--quiet'])

        for arg in bcond_with:
            cmd.extend(['--with', arg])

        for arg in bcond_without:
            cmd.extend(['--without', arg])

        cmd.extend(['--nodeps', '-bs', self.final_spec()])
        run(cmd, capture=False)

    def verify_files(self, target_dir=None):
        """
        Run rpmbuild -bl on the package to verify the %files section.
        """
        if not target_dir:
            self.use_existing_outdir = True
            target_dir = self.outdir

        log.info('Working on {0}...'.format(target_dir))

        cmd = ['rpmbuild'] + self.rpmbuild_opts()

        if self.verbosity == QUIET:
            cmd.extend(['--quiet'])

        try:
            spec_path = glob.glob(os.path.join(target_dir, '*.spec'))[0]
        except IndexError:
            raise RpkgError('No spec file found at {0}'.format(target_dir))

        cmd.extend(['-bl', spec_path])
        run(cmd, capture=False)

    def unused_patches(self):
        final_tmp_spec_path = self.final_tmp_spec()
        unused_patches = glob.glob('*.patch')
        ts = rpm.ts()

        try:
            self.load_rpm_config()
            rpm_spec = ts.parseSpec(final_tmp_spec_path)
        except ValueError as e:
            raise RpkgError("Could not parse spec file with error: "+str(e))
        finally:
            rpm.reloadConfig()
            os.unlink(final_tmp_spec_path)

        for (filename, num, flag) in rpm_spec.sources:
            if flag != 2:
                continue

            if filename in unused_patches:
                unused_patches.remove(filename)

        return unused_patches
